import click
import sys
import os
import shutil

import logging
import csrbox.utils as utils
import csrbox.csr_gen as csr_gen
from   csrbox import __version__ as version
from   csrbox.errors import ValidationError
import riscv_config.checker as riscv_config


@click.command()
@click.version_option(prog_name="RISC-V CSRBox Generator",version=version)
@click.option('--verbose', '-v', default='info', help='Set verbose level', type=click.Choice(['info','error','debug'],case_sensitive=False))

@click.option ('--isaspec','-ispec', type=click.Path(exists=True, resolve_path=True, readable=True), help = "RISCV-CONFIG ISA File" )
@click.option ('--grpspec','-gspec', type=click.Path(resolve_path=True, readable=True, exists=True), help = "Grouping YAML File" )
@click.option ('--customspec','-cspec', type=click.Path(exists=True, resolve_path=True, readable=True), help = "CUSTOM CSR File" )
@click.option ('--debugspec','-dspec', type=click.Path(exists=True, resolve_path=True, readable=True), help = "Debug CSR File" )
@click.option ('--customattr','-cattr', type=click.Path(exists=True, resolve_path=True, readable=True), help = "CUSTOM CSR Attributes File" )
@click.option ('--workdir', default='./csrbox_work', type=click.Path(resolve_path=True, writable=True), help='Work directory path')
@click.option ('--probeinst','-pinst',default='soc', type=str)
def cli(verbose, isaspec, grpspec, customspec, debugspec, customattr, workdir, probeinst):
    utils.setup_logging(verbose)
    logger = logging.getLogger()
    logger.handlers = []
    ch = logging.StreamHandler()
    ch.setFormatter(utils.ColoredFormatter())
    logger.addHandler(ch)

    logger.info('****** RISC-V CSRBox Generator {0} *******'.format(version))
    logger.info('Copyright (c) 2020, IIT-Madras, InCore Semiconductors Pvt. Ltd.')
    if not os.path.exists(workdir):
        logger.debug('Creating new work directory: ' + workdir)
        os.mkdir(workdir)
    else:
        logger.debug('Removing old work directory and creating new one: ' + workdir)
        shutil.rmtree(workdir)
        os.mkdir(workdir)

    try:
        isa_file = riscv_config.check_isa_specs(isaspec, workdir, True)
    except ValidationError as msg:
        logger.error(msg)
        return 1

    if debugspec:
        try:
            debug_file = riscv_config.check_debug_specs(debugspec, isa_file, workdir, True)
        except ValidationError as msg:
            logger.error(msg)
            return 1
    else:
        debug_file = None


    if customspec:
        try:
            custom_file = riscv_config.check_custom_specs(customspec, workdir, True)
        except ValidationError as msg:
            logger.error(msg)
            return 1
    else:
        custom_file = None

    bsv_dir  = os.path.join(workdir, "bsv/")
    if not os.path.exists(bsv_dir):   
        logger.debug('Creating new bsv directory: ' + bsv_dir) 
        os.mkdir(bsv_dir)

    csr_gen.csr_gen(isa_file, grpspec, custom_file, debug_file, customattr, bsv_dir,
            probeinst, logging=True)


 



import os

root = os.path.abspath(os.path.dirname(__file__))
attr_file = os.path.join(root, 'attributes.yaml')

license_header = '''
/*
See LICENSE for details
This file has been generated by CSR-BOX - {0}
Time of Generation: {1}
*/
'''

csr_str_temp = '''
                  `{0}: return "c{1}_{2}";'''

user_trap_csrs = [0x000, 0x004, 0x005, 0x040, 0x041, 0x042, 0x043, 0x044]

standard_csr_fields = [
    'description', 'address', 'priv_mode', 'reset-val', 'implemented', 'msb',
    'lsb', 'type', 'shadow', 'accessible', 'fields'
]

time_csr = ['mtime', 'mtimeh', 'time', 'timeh']

temp_register = '''
    /*doc:reg: {0}*/
    Reg#(Bit#({1})) rg_{2} <- mkReg({3});
'''

temp_shadow_reg = '''
    /*doc:reg: {0}*/
    Reg#(Bit#({1})) rg_{2} = {3};
'''

temp_wire = '''
    /*doc:wire: */
    Wire#(Bit#({0})) rg_{1} <- mkWire();
'''

temp_readOnlyRegister = '''
    /*doc:reg: {0}*/
    Reg#(Bit#({1})) rg_{2} = readOnlyReg({3});
'''

temp_wlrlReg = '''
    function Reg#(Bit#({1})) wlrlReg_{2}(Reg#(Bit#({4})) r);
        return (interface Reg;
            method Bit#({1}) _read = zeroExtend(r);
            method Action _write(Bit#({1}) x);
                r._write(truncate(x));
            endmethod
        endinterface);
    endfunction: wlrlReg_{2}

    /*doc:reg: {0}*/
    Reg#(Bit#({4})) rg_{2}_wlrl <- mkReg({3});
    Reg#(Bit#({1})) rg_{2} = wlrlReg_{2}(rg_{2}_wlrl);
'''

temp_warlReg = '''
    function Reg#(Bit#({1})) warlReg_{2}(Reg#(Bit#({1})) r);
        return (interface Reg;
            method Bit#({1}) _read = r;
            method Action _write(Bit#({1}) x);
                {4}
            endmethod
        endinterface);
    endfunction: warlReg_{2}

    /*doc:reg: {0}*/
    Reg#(Bit#({1})) rg_{2}_warl <- mkReg({3});
    Reg#(Bit#({1})) rg_{2} = warlReg_{2}(rg_{2}_warl);
'''

temp_method_shadow = '''
    method Action ma_read_{1} (Bit#({0}) _{1});
        rg_{1} <= _{1};
    endmethod
'''

temp_req_connection = '''
    mkConnection({0}.mav_fwd_req,{1}.ma_core_req);'''

temp_prv_connection = '''
    mkConnection({0}.ma_upd_privilege, rg_prv);'''

temp_top_csr_connection = '''
    mkConnection({0}.ma_read_{1}, {2}.mv_csr_{1});'''

concat_temp = '''
    /*doc:reg: {0} */
    Reg#(Bit#({1})) rg_{2} = concatReg{3}({4});
'''

readonly_temp = ''' readOnlyReg({0}'d0) '''

address_temp = '''
`define {0} 12'h{1}'''

case_address_temp = '''
                  `{0}: {1} valid= True;'''

package_temp = '''
package {0};
   
import Vector :: *;
import FIFOF :: * ;
import DReg :: * ;
import UniqueWrappers :: * ;
import ConcatReg :: * ;
import GetPut :: * ;
import Connectable :: * ;
import csr_types :: * ;
`include "csrbox.defines"
'''

interface_temp = '''
  // Interaface declaration
  interface Ifc_{0};
'''

methodv_dec_temp = '''
    method Bit#({0}) mv_csr_{1};'''
methoda_read_temp = '''
    method Action ma_read_{1} ({0});'''
method_station_generic = '''    
    /*doc:method : to receive the request from the core or previous node" */
    method Action ma_core_req(CSRReq req); 

    /*doc:method : to send response to core on a hit in this node" */
    method CSRResponse mv_core_resp;

    /*doc:method: fetch from core the prvilege mode */
    method Action ma_upd_privilege (Privilege_mode prv);
    '''

method_fwd_temp = '''
    /*doc:method: to forward the request to the next node on a miss in current node*/
    method ActionValue#(CSRReq) mav_fwd_req;
'''

method_top_def_temp = '''
    method ma_core_req = grp1.ma_core_req;
    method mv_core_resp = CSRResponse{hit:anyhit, data: anydata};
    method mv_prv = rg_prv;'''


method_top_ret_temp = '''
    method ActionValue#(Bit#({0})) mav_upd_on_ret (Privilege_mode prv);
      Bit#(TSub#({0},1)) lv_epc = ?;
'''
method_top_ret_machine_temp = '''
      if (prv == Machine) begin
        {0}.ma_set_mstatus_mpie(1'b1);
        {0}.ma_set_mstatus_mie({0}.mv_csr_mstatus[7]);
        rg_prv <= unpack({0}.mv_csr_mstatus[12:11]);
        lv_epc = truncateLSB({1}.mv_csr_mepc);
        if (lv_misa_u == 1)
          {0}.ma_set_mstatus_mpp(pack(User));
        else
          {0}.ma_set_mstatus_mpp(pack(Machine));
        if (grp1.mv_csr_mstatus[12:11] != '1)
          {0}.ma_set_mstatus_mprv(0);
      end
      if (lv_misa_c == 0)
        lv_epc[0] = 0;
      return {{lv_epc, 1'b0}};
    endmethod
'''
method_top_ret_user_temp = '''
      if (prv == User) begin
        {0}.ma_set_mstatus_upie(1'b1);
        {0}.ma_set_mstatus_uie({0}.mv_csr_mstatus[4]);
        lv_epc = truncateLSB({1}.mv_csr_uepc);
        rg_prv <= User;
        {0}.ma_set_mstatus_mprv(0);
      end
'''
method_top_ret_supervisor_temp = '''
      if (prv == Supervisor) begin
        {0}.ma_set_mstatus_spie(1'b1);
        {0}.ma_set_mstatus_sie({0}.mv_csr_mstatus[5]);
        rg_prv <= unpack(zeroExtend({0}.mv_csr_mstatus[8]));
        lv_epc = truncateLSB({1}.mv_csr_sepc);
        if (lv_misa_u == 1)
          {0}.ma_set_mstatus_spp(0);
        else
          {0}.ma_set_mstatus_spp(1);
        {0}.ma_set_mstatus_mprv(0);
      end
'''
method_top_trap_temp = '''    
    method ActionValue#(Bit#({0})) mav_upd_on_trap(Bit#(6) cause, Bit#({0}) pc, Bit#({0}) tval);
      Bit#(5) lv_cause = truncate(cause);
      Bit#(1) lv_trap_type = truncateLSB(cause);
      Bit#({0}) lv_tvec = ?;
      Privilege_mode prv = Machine;
      Bool delegateM = False;
      Bool delegateS = False;
'''

method_top_trap_medeleg = '''
      let medeleg = {0}.mv_csr_medeleg;
      let mideleg = {1}.mv_csr_mideleg;
      delegateM = (truncate(mideleg >> lv_cause) & lv_trap_type) == 1 ||
                       (truncate(medeleg >> lv_cause) & ~lv_trap_type) == 1;

      if (delegateM && (pack(rg_prv) <= pack(Supervisor)) && (lv_misa_s == 1))
        prv = Supervisor;
      else if (delegateM && delegateS && rg_prv == User && lv_misa_n == 1 && lv_misa_s == 1)
        prv = User;
      else if (delegateM && rg_prv == User && lv_misa_n == 1 && lv_misa_s == 0)
        prv = User;
'''

method_top_trap_sedeleg = '''
      let sedeleg = {0}.mv_csr_sedeleg;
      let sideleg = {1}.mv_csr_sideleg;
      delegateS = (truncate(sideleg >> lv_cause) & lv_trap_type) == 1 ||
                       (truncate(sedeleg >> lv_cause) & ~lv_trap_type) == 1;
'''

method_top_trap_debug_temp = '''
      if ( (lv_trap_type == 0 && (lv_cause == `halt_ebreak || 
                                lv_cause == `halt_trigger ||
                                lv_cause == `halt_step ||
                                lv_cause == `halt_reset ))
            || (lv_trap_type == 1 && lv_cause == `halt_interrupt)) begin
        {0}.ma_set_dpc(pc);
        {1}.ma_set_dcsr_cause(truncate(lv_cause));
        rg_core_halted <= 1;
        {1}.ma_set_dcsr_prv(pack(rg_prv));
        rg_prv <= Machine;
        lv_tvec = {2}.mv_csr_dtvec;
      end
      else if (lv_trap_type == 1 && lv_cause == `resume_interrupt) begin
        lv_tvec = {{truncateLSB({0}.mv_csr_dpc),1'b0}};
        rg_core_halted <= 0;
        rg_prv <= unpack({1}.mv_csr_dcsr[1:0]);
      end
      else
'''

method_top_debug_dec = '''
    method Bit#(1) mv_core_halted;
    method Bit#(1) mv_core_hasreset;
'''
method_top_debug_def = '''
    method mv_core_halted = rg_core_halted;
    method mv_core_hasreset = rg_core_has_reset;
'''

method_top_trap_machine_temp = '''
      if (prv == Machine) begin
        rg_prv <= Machine;
        {0}.ma_set_mtval(tval);
        {1}.ma_set_mepc(pc);
        {2}.ma_set_mcause({{lv_trap_type, 'd0, lv_cause}});
        {3}.ma_set_mstatus_mie(1'b0);
        {3}.ma_set_mstatus_mpp(pack(rg_prv));
        {3}.ma_set_mstatus_mpie({3}.mv_csr_mstatus[3]);
        Bit#(2) lv_trapmode = truncate({4}.mv_csr_mtvec);
        lv_tvec = {{truncateLSB({4}.mv_csr_mtvec),2'b0}};
        if ( lv_trapmode == 1 && lv_trap_type == 1)
            lv_tvec =  lv_tvec + {{zeroExtend(lv_cause),2'b0}};
      end
      return lv_tvec;
    endmethod
'''
method_top_trap_supervisor_temp = '''
      if (prv == Supervisor) begin
        rg_prv <= Supervisor;
        {0}.ma_set_stval(tval);
        {1}.ma_set_sepc(pc);
        {2}.ma_set_scause({{lv_trap_type, 'd0, lv_cause}});
        {3}.ma_set_mstatus_sie(1'b0);
        {3}.ma_set_mstatus_spp(truncate(pack(rg_prv)));
        {3}.ma_set_mstatus_spie({3}.mv_csr_mstatus[1]);
        Bit#(2) lv_trapmode = truncate({4}.mv_csr_stvec);
        lv_tvec = {{truncateLSB({4}.mv_csr_stvec),2'b0}};
        if ( lv_trapmode == 1 && lv_trap_type == 1)
          lv_tvec =  lv_tvec + {{zeroExtend(lv_cause),2'b0}};
      end
      else
'''
method_top_trap_user_temp = '''
      if (prv == User) begin
        rg_prv <= User;
        {0}.ma_set_utval(tval);
        {1}.ma_set_uepc(pc);
        {2}.ma_set_ucause({{lv_trap_type, 'd0, lv_cause}});
        {3}.ma_set_mstatus_uie(1'b0);
        {3}.ma_set_mstatus_upie({3}.mv_csr_mstatus[0]);
        Bit#(2) lv_trapmode = truncate({4}.mv_csr_utvec);
        lv_tvec = {{truncateLSB({4}.mv_csr_stvec),2'b0}};
        if ( lv_trapmode == 1 && lv_trap_type == 1)
          lv_tvec =  lv_tvec + {{zeroExtend(lv_cause),2'b0}};
      end
      else
'''

end_interface_temp = '''
  endinterface
'''
module_temp = '''
  //Module Declarations
  (*synthesize*)
  
  (*conflict_free="ma_core_req,mv_core_resp"*)
{1}
  module mk_{0}{2}(Ifc_{0});
'''

methodv_def_temp = '''
    method mv_csr_{0} = {1}{2};
'''

methoda_top_temp = '''
    method {0} = {1}.{0};'''

end_module_temp = '''
  endmodule '''

endcase_temp = '''
            default: begin
                rg_resp_to_core <= CSRResponse{hit: True, data: 0};
            end
        endcase
        endmethod
    '''
endcase_fwd_temp = '''
            default: begin
                ff_fwd_request.enq(req);
                rg_resp_to_core <= CSRResponse{hit: False, data: 0};
            end
        endcase
        endmethod
    '''

end_package_temp = '''
endpackage 
'''

func_temp = '''
    function Bit#({0}) fn_csr_op (Bit#({0}) writedata, Bit#({0}) readdata, Bit#(2) op);
        if(op == 'd1)
    	    return writedata;
        else if(op == 'd2)
            return (writedata|readdata);
        else
            return (~writedata & readdata);
    endfunction
  '''

func_decoder = '''
    function Bool address_valid(Bit#(12) addr, Bit#(26) misa);
    Bool valid=False;
    case(addr)
   '''

func_csr_to_csr = '''
    function String fn_csr_to_str(Bit#(12) addr);
    case(addr)
    '''

method_csr_req_def_temp = ''' 
        method Action ma_core_req(CSRReq req);
        Bit#(2) op = req.funct3; '''

method_case_req = '''
        case (req.csr_address) '''

method_generic_def_temp = '''
    method mv_core_resp = rg_resp_to_core;

    method Action ma_upd_privilege (Privilege_mode prv);
        wr_prv <= prv;
    endmethod'''

method_fwd_def_temp = '''
        method ActionValue#(CSRReq) mav_fwd_req;
            ff_fwd_request.deq;
            return ff_fwd_request.first();
        endmethod
'''
case_def = '''
            `{0} : begin
                Bit#({2}) readdata = zeroExtend(rg_{1});
                rg_resp_to_core <= CSRResponse{{hit:True, data: readdata}};
                let word = fn_csr_op(req.writedata, readdata, op);
                
                {3} rg_{1} <= truncate(word);
            end
'''
case_def_ro = '''
            `{0} : begin
                Bit#({2}) readdata = zeroExtend(rg_{1});
                rg_resp_to_core <= CSRResponse{{hit:True, data: readdata}};
            end
'''
csrreq_struct_temp = '''
    typedef struct {{
        Bit#(12) csr_address;
        Bit#({0}) writedata;
        Bit#(2) funct3;
        {1}
    }} CSRReq deriving(Bits, FShow, Eq);

    typedef struct{{
        Bool hit;
        Bit#({0})  data;
    }} CSRResponse deriving(Bits, Eq, FShow);

    typedef enum {{Machine = 3, Supervisor = 1, User = 0}} Privilege_mode deriving(Bits, Eq, FShow);

'''

other_vars_temp = '''

    /*doc:wire: holds the response of this group for a csr operation request,
    for one cycle, wire is used for low latency*/
    Wire#(CSRResponse) rg_resp_to_core <- mkDWire(CSRResponse{hit:False, data:0});

    /*doc:fifo: fifo to forward the core request to the next group on a miss*/
    FIFOF#(CSRReq) ff_fwd_request <- mkFIFOF1();

    /*doc:wire: holds the current privilege mode of the hart*/
    Wire#(Privilege_mode) wr_prv <- mkWire();
'''

check_hit_temp = '''
    Bool anyhit = {0};
    Bit#({2}) anydata = {1};
'''

method_top_generic = '''    
    /*doc:method : to receive the request from the core or previous node" */
    method Action ma_core_req(CSRReq req); 

    /*doc:method : to send response to core on a hit in this node" */
    method CSRResponse mv_core_resp;

    method ActionValue#(Bit#({0})) mav_upd_on_ret (Privilege_mode prv);
    
    method ActionValue#(Bit#({0})) mav_upd_on_trap(Bit#(6) cause, Bit#({0}) pc, Bit#({0}) tval);

    method Privilege_mode mv_prv;

'''

pmp_ifc = '''
    method Vector#({0}, Bit#(8)) mv_pmpcfg;
    method Vector#({0}, Bit#({1})) mv_pmpaddr;'''

pmpaddr_def = '''
    method Vector#({0}, Bit#({1})) mv_pmpaddr;
        Vector#({0}, Bit#({1})) lv_pmpaddr;
{2}
        return lv_pmpaddr;
    endmethod:mv_pmpaddr
'''
pmpcfg_def = '''
    method Vector#({0}, Bit#(8)) mv_pmpcfg;
        Vector#({0}, Bit#(8)) lv_pmpcfg;
{1}
        return lv_pmpcfg;
    endmethod:mv_pmpcfg
'''
same_groups_rv32 = [
        ('SSTATUS', 'MSTATUS'),
        ('USTATUS', 'MSTATUS'),
        ('MSTATUS', 'FCSR'),
        ('FRM', 'FCSR'),
        ('FFLAGS', 'FCSR'),
        ('MHPMCOUNTER3', 'MHPMCOUNTER3H'),
        ('MHPMCOUNTER4', 'MHPMCOUNTER4H'),
        ('MHPMCOUNTER5', 'MHPMCOUNTER5H'),
        ('MHPMCOUNTER6', 'MHPMCOUNTER6H'),
        ('MHPMCOUNTER7', 'MHPMCOUNTER7H'),
        ('MHPMCOUNTER8', 'MHPMCOUNTER8H'),
        ('MHPMCOUNTER9', 'MHPMCOUNTER9H'),
        ('MHPMCOUNTER10', 'MHPMCOUNTER10H'),
        ('MHPMCOUNTER11', 'MHPMCOUNTER11H'),
        ('MHPMCOUNTER12', 'MHPMCOUNTER12H'),
        ('MHPMCOUNTER13', 'MHPMCOUNTER13H'),
        ('MHPMCOUNTER14', 'MHPMCOUNTER14H'),
        ('MHPMCOUNTER15', 'MHPMCOUNTER15H'),
        ('MHPMCOUNTER16', 'MHPMCOUNTER16H'),
        ('MHPMCOUNTER17', 'MHPMCOUNTER17H'),
        ('MHPMCOUNTER18', 'MHPMCOUNTER18H'),
        ('MHPMCOUNTER19', 'MHPMCOUNTER19H'),
        ('MHPMCOUNTER20', 'MHPMCOUNTER20H'),
        ('MHPMCOUNTER21', 'MHPMCOUNTER21H'),
        ('MHPMCOUNTER22', 'MHPMCOUNTER22H'),
        ('MHPMCOUNTER23', 'MHPMCOUNTER23H'),
        ('MHPMCOUNTER24', 'MHPMCOUNTER24H'),
        ('MHPMCOUNTER25', 'MHPMCOUNTER25H'),
        ('MHPMCOUNTER26', 'MHPMCOUNTER26H'),
        ('MHPMCOUNTER27', 'MHPMCOUNTER27H'),
        ('MHPMCOUNTER28', 'MHPMCOUNTER28H'),
        ('MHPMCOUNTER29', 'MHPMCOUNTER29H'),
        ('MHPMCOUNTER30', 'MHPMCOUNTER30H'),
        ('MHPMCOUNTER31', 'MHPMCOUNTER31H'),
        ('HPMCOUNTER3', 'HPMCOUNTER3H'),
        ('HPMCOUNTER4', 'HPMCOUNTER4H'),
        ('HPMCOUNTER5', 'HPMCOUNTER5H'),
        ('HPMCOUNTER6', 'HPMCOUNTER6H'),
        ('HPMCOUNTER7', 'HPMCOUNTER7H'),
        ('HPMCOUNTER8', 'HPMCOUNTER8H'),
        ('HPMCOUNTER9', 'HPMCOUNTER9H'),
        ('HPMCOUNTER10', 'HPMCOUNTER10H'),
        ('HPMCOUNTER11', 'HPMCOUNTER11H'),
        ('HPMCOUNTER12', 'HPMCOUNTER12H'),
        ('HPMCOUNTER13', 'HPMCOUNTER13H'),
        ('HPMCOUNTER14', 'HPMCOUNTER14H'),
        ('HPMCOUNTER15', 'HPMCOUNTER15H'),
        ('HPMCOUNTER16', 'HPMCOUNTER16H'),
        ('HPMCOUNTER17', 'HPMCOUNTER17H'),
        ('HPMCOUNTER18', 'HPMCOUNTER18H'),
        ('HPMCOUNTER19', 'HPMCOUNTER19H'),
        ('HPMCOUNTER20', 'HPMCOUNTER20H'),
        ('HPMCOUNTER21', 'HPMCOUNTER21H'),
        ('HPMCOUNTER22', 'HPMCOUNTER22H'),
        ('HPMCOUNTER23', 'HPMCOUNTER23H'),
        ('HPMCOUNTER24', 'HPMCOUNTER24H'),
        ('HPMCOUNTER25', 'HPMCOUNTER25H'),
        ('HPMCOUNTER26', 'HPMCOUNTER26H'),
        ('HPMCOUNTER27', 'HPMCOUNTER27H'),
        ('HPMCOUNTER28', 'HPMCOUNTER28H'),
        ('HPMCOUNTER29', 'HPMCOUNTER29H'),
        ('HPMCOUNTER30', 'HPMCOUNTER30H'),
        ('HPMCOUNTER31', 'HPMCOUNTER31H'),
        ('CYCLE', 'CYCLEH'),
        ('MINSTRET', 'MINSTRETH'),
        ('TIME', 'TIMEH'),
        ('MIE', 'SIE'),
        ('MIE', 'UIE'),
        ('MIP', 'SIP'),
        ('MIP', 'UIP')
        ]

same_groups_rv64 = [
        ('SSTATUS', 'MSTATUS'),
        ('USTATUS', 'MSTATUS'),
        ('MSTATUS', 'FCSR'),
        ('FRM', 'FCSR'),
        ('FFLAGS', 'FCSR'),
        ('MIE', 'SIE'),
        ('MIE', 'UIE'),
        ('MIP', 'SIP'),
        ('MIP', 'UIP'),
        ]

necessary_pairs_rv32 = [
        ('MHPMCOUNTER3', 'MHPMCOUNTER3H'),
        ('MHPMCOUNTER4', 'MHPMCOUNTER4H'),
        ('MHPMCOUNTER5', 'MHPMCOUNTER5H'),
        ('MHPMCOUNTER6', 'MHPMCOUNTER6H'),
        ('MHPMCOUNTER7', 'MHPMCOUNTER7H'),
        ('MHPMCOUNTER8', 'MHPMCOUNTER8H'),
        ('MHPMCOUNTER9', 'MHPMCOUNTER9H'),
        ('MHPMCOUNTER10', 'MHPMCOUNTER10H'),
        ('MHPMCOUNTER11', 'MHPMCOUNTER11H'),
        ('MHPMCOUNTER12', 'MHPMCOUNTER12H'),
        ('MHPMCOUNTER13', 'MHPMCOUNTER13H'),
        ('MHPMCOUNTER14', 'MHPMCOUNTER14H'),
        ('MHPMCOUNTER15', 'MHPMCOUNTER15H'),
        ('MHPMCOUNTER16', 'MHPMCOUNTER16H'),
        ('MHPMCOUNTER17', 'MHPMCOUNTER17H'),
        ('MHPMCOUNTER18', 'MHPMCOUNTER18H'),
        ('MHPMCOUNTER19', 'MHPMCOUNTER19H'),
        ('MHPMCOUNTER20', 'MHPMCOUNTER20H'),
        ('MHPMCOUNTER21', 'MHPMCOUNTER21H'),
        ('MHPMCOUNTER22', 'MHPMCOUNTER22H'),
        ('MHPMCOUNTER23', 'MHPMCOUNTER23H'),
        ('MHPMCOUNTER24', 'MHPMCOUNTER24H'),
        ('MHPMCOUNTER25', 'MHPMCOUNTER25H'),
        ('MHPMCOUNTER26', 'MHPMCOUNTER26H'),
        ('MHPMCOUNTER27', 'MHPMCOUNTER27H'),
        ('MHPMCOUNTER28', 'MHPMCOUNTER28H'),
        ('MHPMCOUNTER29', 'MHPMCOUNTER29H'),
        ('MHPMCOUNTER30', 'MHPMCOUNTER30H'),
        ('MHPMCOUNTER31', 'MHPMCOUNTER31H'),
        ('HPMCOUNTER3', 'HPMCOUNTER3H'),
        ('HPMCOUNTER4', 'HPMCOUNTER4H'),
        ('HPMCOUNTER5', 'HPMCOUNTER5H'),
        ('HPMCOUNTER6', 'HPMCOUNTER6H'),
        ('HPMCOUNTER7', 'HPMCOUNTER7H'),
        ('HPMCOUNTER8', 'HPMCOUNTER8H'),
        ('HPMCOUNTER9', 'HPMCOUNTER9H'),
        ('HPMCOUNTER10', 'HPMCOUNTER10H'),
        ('HPMCOUNTER11', 'HPMCOUNTER11H'),
        ('HPMCOUNTER12', 'HPMCOUNTER12H'),
        ('HPMCOUNTER13', 'HPMCOUNTER13H'),
        ('HPMCOUNTER14', 'HPMCOUNTER14H'),
        ('HPMCOUNTER15', 'HPMCOUNTER15H'),
        ('HPMCOUNTER16', 'HPMCOUNTER16H'),
        ('HPMCOUNTER17', 'HPMCOUNTER17H'),
        ('HPMCOUNTER18', 'HPMCOUNTER18H'),
        ('HPMCOUNTER19', 'HPMCOUNTER19H'),
        ('HPMCOUNTER20', 'HPMCOUNTER20H'),
        ('HPMCOUNTER21', 'HPMCOUNTER21H'),
        ('HPMCOUNTER22', 'HPMCOUNTER22H'),
        ('HPMCOUNTER23', 'HPMCOUNTER23H'),
        ('HPMCOUNTER24', 'HPMCOUNTER24H'),
        ('HPMCOUNTER25', 'HPMCOUNTER25H'),
        ('HPMCOUNTER26', 'HPMCOUNTER26H'),
        ('HPMCOUNTER27', 'HPMCOUNTER27H'),
        ('HPMCOUNTER28', 'HPMCOUNTER28H'),
        ('HPMCOUNTER29', 'HPMCOUNTER29H'),
        ('HPMCOUNTER30', 'HPMCOUNTER30H'),
        ('HPMCOUNTER31', 'HPMCOUNTER31H'),
        ('CYCLE', 'CYCLEH'),
        ('MINSTRET', 'MINSTRETH'),
        ('TIME', 'TIMEH')
        ]



# See LICENSE.incore for details

"""Top-level package for RISC-V CSR Box Generator."""

__author__ = """IIT-Madras, InCore Semiconductors Pvt Ltd"""
__email__ = 'shakti.iitm@gmail.com'
__version__ = '1.3.0'



# See LICENSE.incore for details

"""The setup script."""

import os
import codecs
import pip
from setuptools import setup, find_packages

import csrbox

# Base directory of package
here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()
def read_requires():
    with open(os.path.join(here, "csrbox/requirements.txt"),"r") as reqfile:
        return reqfile.read().splitlines()

#Long Description
with open("README.rst", "r") as fh:
    readme = fh.read()

setup_requirements = [ ]

test_requirements = [ ]

setup(
    name='csrbox',
    version='1.3.0',
    description="RISC-V CSR-Box Generator",
    long_description=readme + '\n\n',
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "License :: OSI Approved :: BSD License",
          "Development Status :: 4 - Beta"
      ],
    url='https://gitlab.com/incoresemi/ex-box/csrbox',
    author="IIT-Madras, InCore Semiconductors Pvt. Ltd.",
    author_email='shakti.iitm@gmail.com',
    license='BSD-3-Clause',
    packages=find_packages(),
    package_dir={'csrbox': 'csrbox'},
    package_data={
        'csrbox': [
            'requirements.txt', 
            'attributes.yaml'
            ]
        },
    install_requires=read_requires(),
    python_requires='>=3.6.0',
      entry_points={
        'console_scripts': [
            'csrbox=csrbox.main:cli',
        ],
    },
    include_package_data=True,
    keywords='csrbox, riscv',
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
)



import logging
import re
import sys
import os
import csrbox
import csrbox.utils as utils
from csrbox.constants import *
import csrbox.csrgrp_check as csrgrp_check
import datetime
import textwrap
import math

version = csrbox.__version__
logger = logging.getLogger(__name__)

def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work_dir,
        probeinst, logging=False):
    '''
        Top level function which validates the sanity of the grouping yaml 
        followed by calls to subfunctions which generated the corresponding 
        bluespec code. This function further assumes that the riscv-config
        isa yaml provided is the validated/normalized version.

        We also update the the size node for the csrs in the attributes YAML
        file based on the csrs that have been instantiated in the riscv-config
        YAML

        :param isa_file: path to the **validated** riscv-config ISA yaml file
        :param grp_file: path to the csr grouping YAML file.
        :param work_dir: path to the directory where the bluespec files should be generated

        :type isa_file: str
        :type grp_file: str
        :type work_dir: str
    '''
    
    isa_yaml = utils.load_yaml(isa_file)
    grp_yaml = utils.load_yaml(grp_file)
    global debug_enabled
    global compressed_enabled
    global supervisor_enabled

    if 'C' in isa_yaml['hart0']['ISA']:
        compressed_enabled = True
    else:
        compressed_enabled = False

    if 'S' in isa_yaml['hart0']['ISA']:
        supervisor_enabled = True
    else:
        supervisor_enabled = False

    if custom_file :
        custom_yaml= utils.load_yaml(custom_file)
        isa_yaml['hart0'].update(custom_yaml['hart0'])
    if debug_file:
        debug_yaml = utils.load_yaml(debug_file)
        isa_yaml['hart0'].update(debug_yaml['hart0'])
        debug_enabled = True

    else:
        debug_enabled = False

    
    csrgrp_check.csrgrp_check(isa_yaml, grp_yaml, logging)
    attr_yaml = csrgrp_check.calculate_size(isa_yaml, custom_attr_file)
    csr_define(isa_yaml , work_dir , 'csrbox.defines', logging)
    csr_decoder(isa_yaml, work_dir , 'csrbox_decoder.bsv', logging)
    csr_types(isa_yaml  , work_dir , 'csr_types.bsv', logging)
    groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging)

def csr_types(isa_yaml, work_dir,  bsv, logging=False):
    '''
        Function to generate the types bluespec file which includes structures
        and functions that are used across the csrbox modules
        
        :param isa_yaml: the riscv-config isa YAML
        :param work_dir: path to the directory where the bluespec files should be generated
        :param bsv: name of the file to be created inside the work_dir

        :type isa_yaml: dict 
        :type work_dir: str
        :type bsv: str

    '''
    if logging: 
        logger.info('Creating File: ' + bsv)
    bsvfile = open(work_dir + bsv, 'w')
    pkg_name = bsv[:-4]
    
    xlen = str(isa_yaml['hart0']['supported_xlen'][0])

    bsvfile.write(license_header.format(version, str(datetime.datetime.now())))
    bsvfile.write('''
package {0} ;
    `include "csrbox.defines"

'''.format(pkg_name))
    if 'C' in isa_yaml['hart0']['ISA'] :
        c_str = 'Bit#(1) pc_1;'
    else:
        c_str = ''
    bsvfile.write(csrreq_struct_temp.format(xlen, c_str))
    bsvfile.write(func_temp.format(xlen))
    bsvfile.write(end_package_temp)
    bsvfile.close()


def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False):
    '''
        Function which co-ordinates the generation of all the bluespec code for
        the individual groups/stations and the top level bluespec module.

        The function first creates a subset of the
        riscv-config isa yaml based on the csrs present in each group and calls
        various sub-functions to generate the corresponding bsv file of the
        individual groups/station.

        Once the station files are generated the csr_top is called which is used
        to generate the top module instantiating and connecting all the
        individual groups/stations.

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type work_dir: str
    '''

    global physical_addr_sz

    rvxlen = 'rv' + str(max(isa_yaml['hart0']['supported_xlen']))
    xlen = str(max(isa_yaml['hart0']['supported_xlen']))

    station_isa = {}
    csr_isa = {}

    not_acc = []
    max_len = len(grp_yaml)

    for grp, elems in grp_yaml.items():
        if elems != None:
          if 'grp'+str(max_len) == grp:
              last = True
          else:
              last = False
          for node in isa_yaml['hart0']:
             if isinstance(isa_yaml['hart0'][node], dict):
                if 'description' in isa_yaml['hart0'][node]:
                    csr_name = node
                    name_str = str(csr_name)
                    if isa_yaml['hart0'][csr_name][rvxlen]['accessible'] == True:
                      if csr_name.upper() in elems:
                          station_isa[csr_name] = isa_yaml['hart0'][node]
                          csr_isa[csr_name] = isa_yaml['hart0'][node]
                    else:
                       not_acc.append(name_str)

        filename = 'csrbox_' + grp + '.bsv'
        bsvstr = ''
        bsvfile = open(work_dir+filename, 'w')
        if logging:
            logger.info('Writing file: ' + filename)
        pkg_name = filename[:-4]
        bsvstr += str(license_header.format(version, str(datetime.datetime.now())))
        bsvstr += (package_temp.format(pkg_name))
        bsvstr = station_interface(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr = station_module(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr = station_regs(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr = station_modbody(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr = station_req(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr = station_methods(station_isa, grp_yaml, attr_yaml, rvxlen,
                pkg_name, last, bsvstr, logging)
        bsvstr += (end_module_temp)
        bsvstr +=(end_package_temp)
        bsvfile.write(bsvstr)
        station_isa.clear()  
    physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
    bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, rvxlen, 
             'csrbox', logging)
    filename = 'csrbox.bsv'
    bsvfile = open(work_dir + filename, 'w')
    bsvfile.write(bsvstr)
    bsvfile.close()

    filename = 'csr_probe.bsv'
    bsvfile = open(work_dir + filename, 'w')
    bsvfile.write(csr_probe_function(csr_isa, attr_yaml, probeinst, logging))
    bsvfile.close()


def csr_probe_function(isa_yaml, attr_yaml,probeinst, logging=False):
    bsvstr = ''
    bsvstr += '''
  function Bit#(XLEN) fn_probe_csr (Bit#(12) csr_addr);
    case (csr_addr)
'''
    if logging:
        logger.info('-- Creating Probe function ')
    for csrname in isa_yaml:
        if attr_yaml[csrname]['value_method']['required']:
            bsvstr += '''
        `{0} : return zeroExtend({2}.sbread.mv_csr_{1});'''.format(csrname.upper(),
                csrname.lower(), probeinst)
    bsvstr += '''
    default: return 0;
    endcase
  endfunction
'''
    return bsvstr
    


def csr_decoder(isa_yaml, work_dir, bsv, logging=False):

    ''' 
        Function to generate the bluespec decoder file indicating if the access
        to the CSR if valid or not
        :param isa_yaml: the riscv-config isa YAML
        :param work_dir: path to the directory where the bluespec files should be generated
        :param bsv: name of the file to be created inside the work_dir

        :type isa_yaml: dict 
        :type work_dir: str
        :type bsv: str
    '''

    rvxlen = 'rv' + str(isa_yaml['hart0']['supported_xlen'][0])

    definefile = open(work_dir + bsv, 'w')
    definefile.write(license_header.format(version, str(datetime.datetime.now())))
    definefile.write("////////////////CSR DECODER////////////////")
    pkg_name = bsv[:-4]
    definefile.write(package_temp.format(pkg_name))
    definefile.write(func_decoder)

    for node in isa_yaml['hart0']:
        cond_str = ''
        if isinstance(isa_yaml['hart0'][node], dict):
            if 'address' in isa_yaml['hart0'][node] and ( isa_yaml['hart0'][node][rvxlen]['accessible'] == True) :
                if 'S' in isa_yaml['hart0'][node]['priv_mode']:
                    cond_str = 'if (misa[18] == 1)'
                elif 'U' in isa_yaml['hart0'][node]['priv_mode'] and \
                    isa_yaml['hart0'][node]['address'] in user_trap_csrs:
                        cond_str = 'if ((misa[13]&misa[20]) == 1)'

                csr_name = node
                definefile.write(
                        case_address_temp.format(csr_name.upper(), cond_str))
    definefile.write('\n           endcase')
    definefile.write('\n     return valid;')
    definefile.write('\n     endfunction')

    definefile.write(func_csr_to_csr)
    for node in isa_yaml['hart0']:
        cond_str = ''
        if isinstance(isa_yaml['hart0'][node], dict):
            if 'address' in isa_yaml['hart0'][node] and ( isa_yaml['hart0'][node][rvxlen]['accessible'] == True) :
                addr = int(isa_yaml['hart0'][node]['address'])
                definefile.write(csr_str_temp.format(node.upper(), addr,
                    node.lower()))
    definefile.write('\n           endcase')
    definefile.write('\n     endfunction')

    definefile.write(end_package_temp)
    definefile.close()

def csr_define(isa_yaml, work_dir, bsv, logging=False):
    '''
        Function which generates the defines file which captues the macro
        encodings used across the csrbox bsv modules

        :param isa_yaml: the riscv-config isa YAML
        :param work_dir: path to the directory where the bluespec files should be generated
        :param bsv: name of the file to be created inside the work_dir

        :type isa_yaml: dict 
        :type work_dir: str
        :type bsv: str
    '''

    if logging:
        logger.info('Creating Defines file: ' + bsv)
    rvxlen = 'rv' + str(isa_yaml['hart0']['supported_xlen'][0])
    definefile = open(work_dir + bsv, 'w')
    definefile.write(license_header.format(version, str(datetime.datetime.now())))
    definefile.write("////////////////CSR LIST////////////////")

    for node in isa_yaml['hart0']:
        if isinstance(isa_yaml['hart0'][node], dict):
            if 'address' in isa_yaml['hart0'][node] and ( isa_yaml['hart0'][node][rvxlen]['accessible'] == True) :
                csr_name = node
                if logging:
                 logger.debug('Processing CSR: ' + str(csr_name.upper()))
                
                csr_address = format(isa_yaml['hart0'][csr_name]['address'],'x')

                if logging:
                 logger.debug('|-- address: ' + str(csr_address))
                 
                definefile.write(address_temp.format(csr_name.upper(),csr_address))

    if isa_yaml['hart0']['custom_exceptions'] is not None:
        for ce in isa_yaml['hart0']['custom_exceptions']:
            definefile.write('\n`define {0} {1}'.format(ce['cause_name'],
                ce['cause_val']))
    if isa_yaml['hart0']['custom_interrupts'] is not None:
        for ce in isa_yaml['hart0']['custom_interrupts']:
            definefile.write('\n`define {0} {1}'.format(ce['cause_name'],
                ce['cause_val']))
    definefile.close()


def create_concat_vector(mylist, xlen):
    '''
    Function to return a bsv-string for a csr which is the concetation of the
    all its subfields and gaps replaced with zeros'

    :param mylist: A list of 3 tuple entries. Each entry has the following
    information per subfield [msb of subfield, width of the subfield, name of
    the subfield]

    :return: string with concatenation of all the sub-fields.
  '''
    concat_length = 0
    newlist = sorted(mylist, key=lambda x: x[0])[::-1]
    cv = ' '
    last_lsb = xlen - 1
    while len(newlist) != 0:
        msb, width, name, type_reg = newlist[0]
        newlist.pop(0)
        if msb != last_lsb:
            cv += readonly_temp.format(last_lsb - msb) + ','
            concat_length += 1
        if 'ro_variable' in type_reg and (name != 'mstatus_sd'):
            cv += 'readOnlyReg(rg_' + str(name) + ' ),'
        else:
            cv += 'rg_' + str(name) + ' ,'
        concat_length += 1
        last_lsb = msb - width

    return concat_length, cv[:-1]


def get_reset_val(value, width, lsb, xlen):
    ''' 
    Function to extract 'width' number of bits starting at the 'lsb' position of
    the value. Note here that the python indexes the binary string from the let
    while the lsb defines the position from the right. Hence the need for this
    function

    :param value: the integer value from which bits need to be extracted.
    :param width: size of extraction
    :param lsb:   starting position of extraction

    :return: returns the reset value in decimal form
  '''
    b = '{:0' + xlen + 'b}'
    # convert number into binary first
    binary = b.format(value)

    # remember python indexes from left to right
    end = len(binary) - lsb - 1
    start = end - width + 1
    kBitSubStr = binary[start:end + 1]

    # convert extracted sub-string into decimal again
    return int(kBitSubStr, 2)

def derive_wlrl_func(wlrl_list):

    legal_list = []
    for l in wlrl_list:
        if ':' in l:
            minval = int(l.split(':')[0],16)
            maxval = int(l.split(':')[1],16)
            legal_list += list(range(minval,maxval+1))

    act_width = int(math.log2(max(legal_list)))

    act_list = list(range(0,(2**act_width)+1))
    if sorted(legal_list) == act_list:
        full_wlrl = True
    else:
        full_wlrl = False

    return full_wlrl, act_width
    

def derive_warl_func(field_name, width, warl, csr_dep):
    '''
        Function to convert the warl strings in the riscv-config YAML to
        corresponding bluespec code.

        The function first splits the warl string into the condition and the
        legalizing strings. A regular expression is then applied to both strings
        separately to extract parameters and create bluespec code.


        :param field_name: string indicating the name of the csr
        :param width: indicates the max size of the csr
        :param warl_str: containts the warl string from the riscv-config YAML string

        :type field_name: str
        :type width: int
        :type warl_str: str
    '''

    bsv_str = ''
    regex_range = re.compile(field_name + '\[(.*?)\]\s*(in|not in)\s*\[(.*?)\]')
    regex_bitmask = re.compile(field_name + '\[(.*?)\]\s*bitmask\s*\[(.*?)\]')
    regex_cond = re.compile('\s*(.*?)\[(.*?)\]\s*(.*?)\s*\[(.*?)\]')
    for warl_str in warl:
        warl_split = warl_str.split('->')
        legal = warl_split[-1]
        if csr_dep:
            condition_str = 'if ('
        else:
            condition_str = ''
        if csr_dep:
            search = regex_cond.findall(warl_split[0])
            for part in search:
                csr, indices, op, val = part
                if field_name == csr:
                    csr_write = 'r'
                else:
                    for dep_csr in csr_dep:
                        if csr in dep_csr and '::' in dep_csr:
                            csr_write = 'rg_' + dep_csr.split('::')[0] + '_'+dep_csr.split('::')[1]
                        elif csr == dep_csr:
                            csr_write = 'rg_' + csr
                op = '!=' if 'not in' in op else '=='
                condition_str += ' {0}[{1}] {2} {3}'.format(csr_write, indices, op, val)
                condition_str += ' && '
            condition_str = condition_str[:-4] + ' )\n                  '
        


        range_vals = regex_range.findall(legal)
        bitmask_vals = regex_bitmask.findall(legal)
        full_warl = False
        curr_bsv_str = ''
        if range_vals != []:
            curr_bsv_str += 'if ( ('
            for x in range_vals:
                curr_bsv_str += '('
                indices = x[0]
                if 'not' in x[1]:
                    not_eq = True
                else:
                    not_eq = False
                values = x[2].split(',')
                for v in values:
                    if ':' in v:
                        base_bound = v.split(':')
                        if int(base_bound[0],16) == 0 and int(base_bound[1],16) == (2**int(width) -1):
                            full_warl = True
                        curr_bsv_str += '( x[' + indices + '] >= ' +\
                        base_bound[0] + ' && x[' + indices + '] <= ' +base_bound[1] + ') ||'
                    elif not_eq:
                        curr_bsv_str += '( x[' + indices + '] != ' + v + ') &&'
                    else:
                        curr_bsv_str += '( x[' + indices + '] == ' + v + ') ||'
                curr_bsv_str = curr_bsv_str[:-2] + ') && '
            curr_bsv_str = curr_bsv_str[:-3] + ')'
            curr_bsv_str = curr_bsv_str.replace('0x', "'h") + ')\n'
            curr_bsv_str += '                        r._write(x);'
        elif bitmask_vals != []:
            indices = bitmask_vals[0][0]
            mask_fixed = bitmask_vals[0][1].split(',')
            mask = mask_fixed[0][2:]
            fixedval = mask_fixed[1][3:]
            curr_bsv_str = "r._write( (x & 'h" + mask + ") | (~'h" + mask + " & " + fixedval + ") );"
        if full_warl:
            curr_bsv_str = 'r._write(x);'
        bsv_str += condition_str + curr_bsv_str + '\n'
        bsv_str += '                else  '
    bsv_str = bsv_str[:-22]
    return bsv_str

def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):

    '''
        Function to create the interface definition of individual
        groups/stations. 

        First Value methods are created for all csrs that are instantiated in
        the particular group. Next, custom action methods required for specific
        csrs of their subfields are created.

        Finally action methods to read shadow or dependent csrs is created.

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''
    global debug_enabled
    global compressed_enabled
    global supervisor_enabled
    xlen = int(rvxlen[2:])
    bsvstr += interface_temp.format(pkg_name)

    if logging:
        logger.info('Initiating interface generation for : ' + pkg_name)
    # define additional methods for certain csrs
    req_ifcs = []
    for csrname in isa_yaml:
        if csrname.lower() in attr_yaml and 'additional_method' in attr_yaml[csrname.lower()]:
            for add_ifc in attr_yaml[csrname.lower()]['additional_method']:
                if '::' in add_ifc:
                    cond = add_ifc.split('::')[0]
                    if cond:
                        add_ifc = add_ifc.split('::')[1]
                    else:
                        add_ifc = ''
                if add_ifc not in req_ifcs:
                    bsvstr +=\
                        textwrap.indent(add_ifc, '    ') + '\n'
                    req_ifcs.append(add_ifc)

    # define a value method for each csr.
    for csrname in isa_yaml:
        if attr_yaml[csrname]['value_method']['required']:
            if  logging:
                logger.debug('-- Creating value methods for ' + csrname)
            size = attr_yaml[csrname]['size']
            bsvstr +=(methodv_dec_temp.format(size, csrname))

    # define csr specific action methods.
    for csrname in isa_yaml:
        if attr_yaml[csrname]['action_method']['required']:
            if  logging:
                logger.debug('-- Creating action methods for ' + csrname)
            size = attr_yaml[csrname]['size']
            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
                arguments = arguments.replace('xlen',str(xlen))
                arguments = arguments.replace('size',str(size))
                if '::' in arguments:
                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
                    if eval(cond):         
                        args = textwrap.indent(args, '')
                    else:
                        args = ''
                else:
                    args = arguments
                bsvstr +=('\n    '+args)
    bsvstr +=('\n')

    # create action methods of dependent csrs
    incomming_reads = []
    for csrname in isa_yaml:
        if 'depends_on_csr' in attr_yaml[csrname]:
            if len(attr_yaml[csrname]['depends_on_csr']) != 0:
                if  logging:
                    logger.debug('-- Creating action/read methods for dependencies of ' + csrname)
                for dependent in attr_yaml[csrname]['depends_on_csr']:
                    create_sideband = True
                    dcsr = dependent
                    if '::' in dependent:
                        condition, csr = re.findall(r'^(.*?)::(.*?)$',dependent)[0]
                        if eval(condition) :
                            dcsr = csr
                        else:
                            create_sideband = False
                    if create_sideband and dcsr not in isa_yaml and \
                            dcsr not in incomming_reads \
                            and attr_yaml[dcsr]['size'] is not None:
                        size = attr_yaml[dcsr]['size']
                        bsvstr +=(methoda_read_temp.format('Bit#('+str(size)
                            + ') _' + dcsr, dcsr))
                        incomming_reads.append(dcsr)

    # create methods for shadow csrs not in the current station
    for csrname, content in isa_yaml.items():
        subfields = []
        for s in content[rvxlen]['fields']:
            if isinstance(s, list):
                continue
            else:
                subfields.append(s)
        if not subfields:
            size = attr_yaml[csrname]['size']
            if content[rvxlen]['shadow'] is not None:
                if  logging:
                    logger.debug('-- Creating action/read methods for shadow csr :' + csrname)
                shadow_csr = content[rvxlen]['shadow'].split('.')[0]
                if shadow_csr not in isa_yaml and \
                        shadow_csr not in incomming_reads and\
                        attr_yaml[shadow_csr]['size']is not None:
                    bsvstr +=(methoda_read_temp.format('Bit#('+str(size)
                            + ') _' + shadow_csr, shadow_csr))
                    incomming_reads.append(shadow_csr)

    if logging:
        logger.info('Interface created for : ' + pkg_name)
    # define generic methods for all stations
    bsvstr +=(method_station_generic)
    if not last:
        bsvstr +=(method_fwd_temp)
    bsvstr += (end_interface_temp)
    return bsvstr

def station_module(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):

    ''' 
        Function to create the bluespec module syntax and other synthesis 
        attributes of the module 

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''

    if logging:
        logger.info('Creating module for : ' + pkg_name)
    attr_str = ''
    for csrname in isa_yaml:
        if 'mod_attributes' in attr_yaml[csrname]:
            for modattr in attr_yaml[csrname]['mod_attributes']:
                attr_str += '  ' + modattr + '\n'
    bsvstr += (module_temp.format(pkg_name, attr_str, ''))
    if logging:
        logger.info('Module created for : ' + pkg_name)

    return  bsvstr

def station_modbody (isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):
    ''' 
        Function to create the bluespec module rules/body requirements per csr
        basis

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''

    if logging:
        logger.info('Creating custom rules/module elements for : ' + pkg_name)
    for csrname in isa_yaml:
        if 'module_body' in attr_yaml[csrname]:
            bsvstr += textwrap.indent(attr_yaml[csrname]['module_body'],'    ') + '\n'
    if logging:
        logger.info('Custom rules/module elements created for : ' + pkg_name)
    return bsvstr

def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):
    ''' 
        Function to create the bluespec method definitions for each
        group/station. First the value methods of each csr in the group is
        created. Followed by custom action method and then the shadow/dependent
        csrs.

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''
    xlen = int(rvxlen[2:])
    if logging:
        logger.info('Creating method definitions for : ' + pkg_name)
    
    # default value methods
    for csrname in isa_yaml:
        if logging:
            logger.debug('-- Defining value methods for : ' + csrname)
        if attr_yaml[csrname]['value_method']['required']:
            if 'dec_body' in attr_yaml[csrname]['value_method']:
                for dec in attr_yaml[csrname]['value_method']['dec_body']:
                    bsvstr += methodv_def_temp.format(csrname, '', dec)
            else:
                bsvstr += methodv_def_temp.format(csrname, 'rg_', csrname)
    
    # custom action methods
    for csrname in isa_yaml:
        if logging:
            logger.debug('-- Defining custom Action methods for : ' + csrname)
        if attr_yaml[csrname]['action_method']['required']:
            size = attr_yaml[csrname]['size']
            for _str in attr_yaml[csrname]['action_method']['dec_body']:
                _str = _str.replace('size', str(size))
                _str = _str.replace('xlen', str(xlen))
                if '::' in _str:
                    cond = _str.split('::')[0]
                    if eval(cond):
                        _str = _str.split('::')[1]
                    else:
                        _str = ''
                bsvstr += textwrap.indent( _str+ '\n', '    ')
    
    # create action methods of dependent csrs
    incomming_reads = []
    for csrname in isa_yaml:
        if 'depends_on_csr' in attr_yaml[csrname]:
            if len(attr_yaml[csrname]['depends_on_csr']) != 0:
                if logging:
                    logger.debug('-- Defining action methods for dependencies of : ' + csrname)
                for dependent in attr_yaml[csrname]['depends_on_csr']:
                    create_sideband = True
                    dcsr = dependent
                    if '::' in dependent:
                        condition, csr = re.findall(r'^(.*?)::(.*?)$',dependent)[0]
                        if eval(condition) :
                            dcsr = csr
                        else:
                            create_sideband = False
                    if create_sideband and dcsr not in isa_yaml and \
                            dcsr not in incomming_reads \
                            and attr_yaml[dcsr]['size'] is not None:
                        size = attr_yaml[dcsr]['size']
                        if 'reg_instantiation' in attr_yaml[csrname]:
                            if len(attr_yaml[csrname]['reg_instantiation']) !=0 :
                                bsvstr += temp_method_shadow.format(str(size), dcsr)
                        else:
                            bsvstr += temp_method_shadow.format(str(size), dcsr)
                        incomming_reads.append(dcsr)
    # create methods for shadow csrs not in the current station
    for csrname, content in isa_yaml.items():
        subfields = []
        for s in content[rvxlen]['fields']:
            if isinstance(s, list):
                continue
            else:
                subfields.append(s)
        if not subfields:
            size = attr_yaml[csrname]['size']
            if content[rvxlen]['shadow'] is not None:
                if logging:
                    logger.debug('-- Defining Action methods for shadows of : ' + csrname)
                shadow_csr = content[rvxlen]['shadow'].split('.')[0]
                if shadow_csr not in isa_yaml and\
                        attr_yaml[shadow_csr]['size']is not None and \
                        shadow_csr not in incomming_reads:
                    if 'reg_instantiation' in attr_yaml[csrname]:
                        if len(attr_yaml[csrname]['reg_instantiation']) !=0 :
                            bsvstr += temp_method_shadow.format(str(size), shadow_csr)
                    else:
                        bsvstr += temp_method_shadow.format(str(size), shadow_csr)
                    incomming_reads.append(shadow_csr)

    bsvstr += method_generic_def_temp
    if not last:
        bsvstr += method_fwd_def_temp
    return bsvstr

def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):
    ''' 
        The function to create the final methods which serves the request from
        the core for each csr.

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''
    global compressed_enabled

    if logging:
        logger.info('Generating Core request routine for : ' + pkg_name)
    bsvstr += method_csr_req_def_temp
    bsvstr += method_case_req
    xlen = int(rvxlen[2:])

    for csrname in isa_yaml:
        size = attr_yaml[csrname]['size']
        address = isa_yaml[csrname]['address']
        ro = False
        is_shadow = False
        shadow_csr = None
        regname = csrname

        if not isa_yaml[csrname][rvxlen]['fields'] :
            if isa_yaml[csrname][rvxlen]['shadow'] is not None:
                is_shadow = True
                shadow_csr = isa_yaml[csrname][rvxlen]['shadow']

        if (address & 0xC00) == 0xC00:
            ro = True

        if 'write_condition' in attr_yaml[csrname]:
            write_cond = attr_yaml[csrname]['write_condition']
        else:
            write_cond = ''

        if is_shadow: 
            regname = shadow_csr
        logger.debug('Case statement for ' +csrname + ' Address: ' +\
                str(address))

        if 'csr_op' in attr_yaml[csrname] and (\
                shadow_csr is None or shadow_csr in isa_yaml):
                if logging:
                    logger.debug('-- Creating custom logic for : ' + csrname)
                for csrop in attr_yaml[csrname]['csr_op']:
                    if '::' in csrop:
                        cond = csrop.split('::')[0]
                        if eval(cond):
                            cust_csrop = csrop.split('::')[1]
                        else:
                            cust_csrop = ''
                    else:
                        cust_csrop = csrop
                _t = cust_csrop.replace('size',str(size))
                _t = _t.replace('xlen',str(xlen))
                _t = textwrap.indent(_t, '            ') + '\n'
                bsvstr += _t
        elif ro:
            bsvstr += case_def_ro.format(csrname.upper(), regname, xlen)
            if logging:
                logger.debug('-- Creating read-only logic for : ' + csrname)
        else:
            bsvstr += case_def.format(csrname.upper(), regname, xlen, write_cond)
            if logging:
                logger.debug('-- Creating default logic for : ' + csrname)
    
    if last:
        bsvstr += (endcase_temp)
    else:
        bsvstr += (endcase_fwd_temp)
    if logging:
        logger.info('Core request routine created for : ' + pkg_name)

    return bsvstr

def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
        last, bsvstr, logging = False):
    '''
        Function to create the register and wires required to create the csrs.
        The function also creates WARL bluespec syntax using the
        derive_warl_func function.

        First the shadow/dependent wires are declared followed by the
        instantiation of the csrs present in the current group

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkg_name: name of the bluespec package this code belongs to.
        :param last: indicating this is the last station in the chain
        :param bsvstr: string containing rest of the code generated for the particular group/station
        :param work_dir: path to the directory where the bluespec files should be generated
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkg_name: str
        :type last: bool
        :type bsvstr: str
        :type work_dir: str
    '''
    if logging:
        logger.info('Generating storage elements for : ' + pkg_name)
    bsvstr += (other_vars_temp)
    shadows = []
    xlen = int(rvxlen[2:])
    # creating additional required registers
    req_regs = []
    for csrname in isa_yaml:
        if csrname in attr_yaml and 'additional_regs' in attr_yaml[csrname.lower()]:
            for add_reg in attr_yaml[csrname.lower()]['additional_regs']:
                if '::' in add_reg:
                    cond = add_reg.split('::')[0]
                    if eval(cond):
                        add_reg = add_reg.split('::')[1]
                    else:
                        add_reg = ''
                if add_reg not in req_regs:
                    bsvstr += '\n'+textwrap.indent(add_reg,'    ')
                    req_regs.append(add_reg)
#        if csrname.upper() in additional_regs:
#            if additional_regs[csrname.upper()] not in req_regs:
#                bsvstr += textwrap.indent(additional_regs[csrname.upper()],'    ')
#                req_regs.append(additional_regs[csrname.upper()])

    # create wires of dependent csrs
    incomming_reads = []
    for csrname in isa_yaml:
        if 'depends_on_csr' in attr_yaml[csrname]:
            if len(attr_yaml[csrname]['depends_on_csr']) != 0:
                for dependent in attr_yaml[csrname]['depends_on_csr']:
                    create_sideband = True
                    dcsr = dependent
                    if '::' in dependent:
                        condition, csr = re.findall(r'^(.*?)::(.*?)$',dependent)[0]
                        if eval(condition) :
                            dcsr = csr
                        else:
                            create_sideband = False
                    if create_sideband and dcsr not in isa_yaml and \
                            dcsr not in incomming_reads \
                            and attr_yaml[dcsr]['size'] is not None:
                        size = attr_yaml[dcsr]['size']
                        bsvstr += temp_wire.format(size, dcsr)
                        incomming_reads.append(dcsr)


    for csrname,content in isa_yaml.items():
        if logging:
            logger.debug('-- Creating storage for : ' + csrname)
        subfields = []
        for s in content[rvxlen]['fields']:
            if isinstance(s, list):
                continue
            else:
                subfields.append(s)
        description = content['description']
        reset_val = content['reset-val']
        size = attr_yaml[csrname]['size']
        concat = []

        if not subfields:
            if 'reg_instantiation' in attr_yaml[csrname]:
                if len(attr_yaml[csrname]['reg_instantiation']) !=0 :
                    bsvstr += attr_yaml[csrname]['reg_instantiation'] + '\n'
            elif content[rvxlen]['shadow'] is not None:
                shadow_csr = content[rvxlen]['shadow'].split('.')[0]
                if shadow_csr not in isa_yaml and shadow_csr not in incomming_reads and\
                        attr_yaml[shadow_csr]['size']is not None:
                    bsvstr += temp_wire.format(size, shadow_csr)
                    incomming_reads.append(shadow_csr)
            elif 'ro_constant' in content[rvxlen]['type']:
                bsvstr += temp_readOnlyRegister.format(description, str(size),
                        csrname, reset_val)
            elif ('warl' in content[rvxlen]['type']):
                bsv_str = derive_warl_func( csrname, str(size),
                        content[rvxlen]['type']['warl']['legal'], 
                        content[rvxlen]['type']['warl']['dependency_fields'])
                bsvstr += temp_warlReg.format(description, str(size), 
                            csrname, reset_val, bsv_str)
            else:
                bsvstr +=  temp_register.format(description, str(xlen), csrname,
                        reset_val)
        else:
            for subfield in subfields:
                if logging:
                    logger.debug('|---- Subfield: ' + subfield)
                if subfield is not 'wlrl':
                    subcontent = isa_yaml[csrname][rvxlen][subfield]
                    if 'type' in subcontent:
                        type_reg = subcontent['type']
                    else:
                        type_reg = ''
                    msb = subcontent['msb']
                    lsb = subcontent['lsb']
                    subdescription = subcontent['description']
                    width = msb - lsb + 1
                    subreset_val = get_reset_val(reset_val, width, lsb, str(xlen))
                    if subcontent['shadow'] is not None:
                        concat.append([ msb, width, subcontent['shadow'].replace('.','_'), ''])
                    else:
                        concat.append([ msb, width, csrname + '_' + subfield, 
                         type_reg])

                    if subfield == 'sd' and csrname == 'mstatus':
                        bsvstr += temp_readOnlyRegister.format(subdescription,
                                str(width), csrname+'_'+subfield, 
                                "pack((rg_mstatus_xs == 2\'b11) || (rg_mstatus_fs == 2\'b11))")
                    elif subcontent['shadow'] is not None:
                        shadow_csr = subcontent['shadow'].split('.')[0]
                        if shadow_csr not in isa_yaml and shadow_csr not in incomming_reads and\
                                attr_yaml[shadow_csr]['size']is not None:
                            bsvstr += temp_wire.format(size, shadow_csr)
                            incomming_reads.append(shadow_csr)

                    elif 'ro_constant' in type_reg:
                        bsvstr += temp_readOnlyRegister.format(subdescription,
                                str(width), csrname+'_'+subfield, subreset_val)
                    elif 'ro_variable' in type_reg:
                        bsvstr += temp_register.format(subdescription,
                                str(width), csrname+'_'+subfield, subreset_val)
                    elif 'warl' in type_reg:
                        bsv_str = derive_warl_func( subfield,
                            str(width),subcontent['type']['warl']['legal'],
                            subcontent['type']['warl']['dependency_fields'])
                        bsvstr += temp_warlReg.format(subdescription,
                                    str(width), csrname+'_'+subfield, subreset_val, bsv_str)
                    elif 'wlrl' in type_reg:
                        full_wlrl, awidth = derive_wlrl_func(subcontent['type']['wlrl'])
                        if full_wlrl:
                            bsvstr +=  temp_register.format(subdescription,
                                str(width), csrname+'_'+subfield, subreset_val)
                        else:
                            bsvstr += temp_wlrlReg.format(subdescription,
                                    str(width), csrname+'_'+subfield, \
                                            subreset_val, awidth)
                    else:
                        bsvstr += temp_readOnlyRegister.format(subdescription,
                                str(width), csrname+'_'+subfield, 0)
            length, cv = create_concat_vector(concat, int(xlen))
            bsvstr += concat_temp.format(description, xlen, csrname,\
                      length, cv)
    if logging:
        logger.info('Storage elements created for : ' + pkg_name)
    return bsvstr

def find_group(bar, csrname, logging=False):
    '''
        Function to find the group which a particular csr belongs to in the
        grouping yaml

        :param bar: the grouping YAML
        :param csrname: the name of the csr
        :type bar: dict
        :type csrname: str

        :return: name of thr group where the csr belongs. None if it is absent altogether
    '''
    for grps in bar:
        if csrname.upper() in bar[grps]:
            return grps
    return None

def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
    '''
        Function to create the top level bluespec module which instantiates the
        individual group/station and creates the connection between them.

        The function also creates the methods required for handling traps and
        returns.

        :param isa_yaml: the riscv-config isa YAML
        :param grp_yaml: the grouping YAML
        :param attr_yaml: the attributes YAML
        :param rvxlen: a string containing the format: "rv32" or "rv64" depending on the XLEN value of the DUT
        :param pkgname: name of the bluespec package this code belongs to.
        
        :type isa_yaml: dict 
        :type grp_yaml: dict 
        :type attr_yaml: dict 
        :type rvxlen: str
        :type pkgname: str
    '''
    global debug_enabled
    global compressed_enabled

    grp_list = sorted(list(grp_yaml))
    xlen = int(rvxlen[2:])
    if logging:
        logger.info('Creating file Top file : ' + str(pkgname))
    bsvstr = ''

    # initial license and imports
    bsvstr += license_header.format(version, str(datetime.datetime.now()))
    bsvstr += package_temp.format(pkgname)

    misa_grp = find_group(grp_yaml, 'misa')
    mstatus_grp = find_group(grp_yaml, 'mstatus')
    mepc_grp = find_group(grp_yaml, 'mepc')
    mtval_grp = find_group(grp_yaml, 'mtval')
    mcause_grp = find_group(grp_yaml, 'mcause')
    mtvec_grp = find_group(grp_yaml, 'mtvec')
    sstatus_grp = find_group(grp_yaml, 'sstatus')
    sepc_grp = find_group(grp_yaml, 'sepc')
    stval_grp = find_group(grp_yaml, 'stval')
    scause_grp = find_group(grp_yaml, 'scause')
    stvec_grp = find_group(grp_yaml, 'stvec')
    ustatus_grp = find_group(grp_yaml, 'ustatus')
    uepc_grp = find_group(grp_yaml, 'uepc')
    utval_grp = find_group(grp_yaml, 'utval')
    ucause_grp = find_group(grp_yaml, 'ucause')
    utvec_grp = find_group(grp_yaml, 'utvec')
    medeleg_grp = find_group(grp_yaml, 'medeleg')
    mideleg_grp = find_group(grp_yaml, 'mideleg')
    sedeleg_grp = find_group(grp_yaml, 'sedeleg')
    sideleg_grp = find_group(grp_yaml, 'sideleg')
    dpc_grp = find_group(grp_yaml,'dpc')
    dcsr_grp_name = find_group(grp_yaml,'dcsr')
    dtvec_grp = find_group(grp_yaml,'dtvec')

    if debug_enabled:
        if dpc_grp is None or dcsr_grp_name is None or dtvec_grp is None:
            logger.error('Debug is enabled but one of dpc, dcsr and/or dtvec is not available in any group')
            sys.exit(1)
    
    pmp_entries = 0


    usertraps = False
    if None not in (ustatus_grp, uepc_grp, utval_grp, ucause_grp, utval_grp):
        usertraps = True
    supervisor = False
    if None not in (sstatus_grp, sepc_grp, stval_grp, scause_grp, stval_grp):
        supervisor = True



    # imports of each station
    if logging:
        logger.debug('-- Creating imports')
    for groups in grp_yaml:
        bsvstr += 'import csrbox_' + groups + ' :: * ;\n'
    bsvstr += 'import csrbox_decoder :: * ;'
    
    
    if logging:
        logger.debug('-- Creating interface')
    # define a value method for each csr.
    bsvstr += '''
  interface Sbread;'''
    for csrname in isa_yaml:
        if logging:
            logger.debug('-- Creating Interface for ' + csrname)
        if attr_yaml[csrname]['value_method']['required']:
            size = attr_yaml[csrname]['size']
            bsvstr +=(methodv_dec_temp.format(size, csrname))
    bsvstr += '''  
  endinterface:Sbread'''



    bsvstr += interface_temp.format(pkgname)
    bsvstr += '''    interface Sbread sbread;\n'''
    # define additional required interfaces
    req_ifcs = []
    for csrname in isa_yaml:
        if csrname.lower() in attr_yaml and 'additional_top_method' in attr_yaml[csrname.lower()]:
            for add_ifc in attr_yaml[csrname.lower()]['additional_top_method']:
                if '::' in add_ifc:
                    cond = add_ifc.split('::')[0]
                    if cond:
                        add_ifc = add_ifc.split('::')[1]
                    else:
                        add_ifc = ''
                if add_ifc not in req_ifcs:
                    bsvstr +=\
                        textwrap.indent(add_ifc, '    ') + '\n'
                    req_ifcs.append(add_ifc)
        if 'pmpaddr' in csrname.lower():
            if isa_yaml[csrname]['rv'+str(xlen)]['accessible']:
                pmp_entries += 1

    # pmp methods if required
    if pmp_entries > 0:
        bsvstr += pmp_ifc.format(pmp_entries, physical_addr_sz)

    # define csr specific action methods.
    for csrname in isa_yaml:
        if logging:
            logger.debug('-- Creating custom action methods for : ' + csrname)
        if attr_yaml[csrname]['action_method']['required'] and \
                attr_yaml[csrname]['action_method']['global']: 
            size = attr_yaml[csrname]['size']
            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
                arguments = arguments.replace('xlen',str(xlen))
                arguments = arguments.replace('size',str(size))
                if '::' in arguments:
                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
                    if eval(cond):         
                        args = textwrap.indent(args, '')
                    else:
                        args = ''
                else:
                    args = arguments
                bsvstr +=('\n    '+args)

    # debug specific interfaces
    if debug_enabled:
        bsvstr += method_top_debug_dec;
    bsvstr += ('\n')
    bsvstr += (method_top_generic.format(xlen))
    bsvstr += (end_interface_temp)
    # -------------- interface definition done ------------------------------#
    
    #--- create module
    if logging:
        logger.debug('-- Creating module')
    mod_attr_str = ''
    if 'minstret' in isa_yaml:
        if isa_yaml['minstret']['rv'+str(xlen)]['accessible']:
            mod_attr_str = '(*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
    if debug_enabled:
        extra_reset = '#(Reset debug_reset)'
    else:
        extra_reset = ''
    bsvstr += module_temp.format(pkgname,mod_attr_str, extra_reset) + '\n'

    if logging:
        logger.debug('-- Instatiating groups')
    # -- instantiate groups
    for grps in grp_yaml:
        bsvstr += '    let ' + grps + ' <- mk_csrbox_' + grps + ';\n'
   
    bsvstr += '''
    let lv_misa_s = {0}.mv_csr_misa[18];
    let lv_misa_u = {0}.mv_csr_misa[20];
    let lv_misa_n = {0}.mv_csr_misa[13];
    let lv_misa_c = {0}.mv_csr_misa[2];
    /*doc:reg: holds the current privilege level*/
    Reg#(Privilege_mode) rg_prv <- mkReg(Machine);
'''.format(misa_grp)
    
    if debug_enabled:
        bsvstr += '''
    /*doc:reg:The register indicates that the core is halted*/
    Reg#(Bit#(1)) rg_core_halted <- mkReg(0);

    /*doc:reg: this register indicates if the core has been reset by the
    debugger*/
    Reg#(Bit#(1)) rg_core_has_reset <- mkRegA(0, reset_by debug_reset);
'''

    anyhit = ''
    anydata = ''
    for grp in grp_yaml:
        anyhit += grp + '.mv_core_resp.hit' + ' || '
        anydata += grp + '.mv_core_resp.data' + ' | '
    anyhit = anyhit[:-4]
    anydata = anydata[:-3]
    bsvstr += check_hit_temp.format(anyhit, anydata, xlen)
    
    # additional wires/regs
    req_regs = []
    for csrname in isa_yaml:
        if csrname.lower() in attr_yaml and 'additional_top_regs' in attr_yaml[csrname.lower()]:
            grp_name = find_group(grp_yaml, csrname.upper())
            for add_reg in attr_yaml[csrname.lower()]['additional_top_regs']:
                add_reg = add_reg.replace('grp_name',str(grp_name))
                if '::' in add_reg:
                    cond = add_reg.split('::')[0]
                    if cond:
                        add_reg = add_reg.split('::')[1]
                    else:
                        add_reg = ''
                if add_reg not in req_regs and add_reg != '' :
                    bsvstr +=\
                        textwrap.indent(add_reg, '    ') + '\n'
                    req_regs.append(add_reg)

    # req=response connection across stations
    if logging:
        logger.debug('-- Creating connections between core-fwd interfaces')
    for i in range(len(grp_list)-1):
        bsvstr += temp_req_connection.format(grp_list[i], grp_list[i+1])
    
    # -- connect rg_prv to all stations
    for grp in grp_yaml:
        bsvstr += temp_prv_connection.format(grp)

    if debug_enabled:
        bsvstr += '''
    rule rl_core_reset;
        rg_core_has_reset <= 1;
    endrule: rl_core_reset
'''
    
    # create action methods of dependent csrs
    if logging:
        logger.debug('-- defining action methods')
    for grp in grp_yaml:
        incomming_reads = []
        if grp_yaml[grp] != None :
         for csrs in grp_yaml[grp]:
            csrname = csrs.lower()
            if 'depends_on_csr' in attr_yaml[csrname]:
                if len(attr_yaml[csrname]['depends_on_csr']) != 0:
                    for dependent in attr_yaml[csrname]['depends_on_csr']:
                        create_sideband = True
                        dcsr = dependent
                        if '::' in dependent:
                            condition, csr = re.findall(r'^(.*?)::(.*?)$',dependent)[0]
                            if eval(condition) :
                                dcsr = csr
                            else:
                                create_sideband = False
                        if create_sideband and dcsr.upper() not in grp_yaml[grp] and \
                                dcsr not in incomming_reads \
                                and attr_yaml[dcsr]['size'] is not None:
                            if logging:
                                logger.debug('-- Creating connections for dependencies\
 of csr: ' +csrname + ' : ' +  str(dcsr))
                            dcsr_grp = find_group(grp_yaml, dcsr)
                            bsvstr += temp_top_csr_connection.format(\
                                    grp, dcsr, dcsr_grp )
                            incomming_reads.append(dcsr)
            subfields = []
            for s in isa_yaml[csrname][rvxlen]['fields']:
                if isinstance(s, list):
                    continue
                else:
                    subfields.append(s)
            if not subfields:
                size = attr_yaml[csrname]['size']
                if isa_yaml[csrname][rvxlen]['shadow'] is not None:
                    shadow_csr = isa_yaml[csrname][rvxlen]['shadow'].split('.')[0]
                    if shadow_csr.upper() not in grp_yaml[grp] and \
                            shadow_csr not in incomming_reads and\
                            attr_yaml[shadow_csr]['size']is not None:
                        if logging:
                            logger.debug('-- Creating connections for shadows of csr: ' +csrname + ' : ' + shadow_csr)
                        shadow_grp = find_group(grp_yaml, shadow_csr)
                        bsvstr += temp_top_csr_connection.format(\
                                    grp, shadow_csr, shadow_grp )
                        incomming_reads.append(shadow_csr)
            
    # return and trap method definitions
    if logging:
        logger.debug('-- Defining trap and return methods')
    bsvstr += method_top_def_temp
    bsvstr += method_top_ret_temp.format(xlen)
    if usertraps :
        bsvstr += method_top_ret_user_temp.format(mstatus_grp, uepc_grp)
    if supervisor:
        bsvstr += method_top_ret_supervisor_temp.format(mstatus_grp, sepc_grp)
    bsvstr += method_top_ret_machine_temp.format(mstatus_grp, mepc_grp)

    bsvstr += method_top_trap_temp.format(xlen)
    if usertraps and supervisor and medeleg_grp is not None \
            and mideleg_grp is not None:
        bsvstr += method_top_trap_sedeleg.format(sedeleg_grp, sideleg_grp)
    if (usertraps or supervisor) and medeleg_grp is not None:
        bsvstr += method_top_trap_medeleg.format(medeleg_grp, mideleg_grp)
    if debug_enabled:
        bsvstr += method_top_trap_debug_temp.format(dpc_grp, dcsr_grp_name,
                dtvec_grp)
    if supervisor:
        bsvstr += method_top_trap_supervisor_temp.format(stval_grp, sepc_grp, scause_grp,
            mstatus_grp, stvec_grp)
    if usertraps:
        bsvstr += method_top_trap_user_temp.format(utval_grp, uepc_grp, ucause_grp,
            mstatus_grp, utvec_grp)
    bsvstr += method_top_trap_machine_temp.format(mtval_grp, mepc_grp, mcause_grp,
            mstatus_grp, mtvec_grp)
    
    # define additional required interfaces
    req_ifc = []
    for csrname in isa_yaml:
        if csrname.lower() in attr_yaml and 'additional_top_method_body' in attr_yaml[csrname.lower()]:
            for add_ifc in attr_yaml[csrname.lower()]['additional_top_method_body']:
                if '::' in add_ifc:
                    cond = add_ifc.split('::')[0]
                    if cond:
                        add_ifc = add_ifc.split('::')[1]
                    else:
                        add_ifc = ''
                if add_ifc not in req_ifc:
                    bsvstr +=\
                        textwrap.indent(add_ifc, '    ') + '\n'
                    req_ifc.append(add_ifc)

    # pmp method definitions
    pmpaddr_str = ''
    pmpcfg_str = ''
    cfg_count = 0
    count = 0
    if pmp_entries > 0:
        for csrname in isa_yaml:
            if 'pmpaddr' in csrname.lower():
                num = int(re.findall('\d+',csrname.lower())[0])
                grp = find_group(grp_yaml, csrname.upper())
                pmpaddr_str += "        lv_pmpaddr[{0}] = {{truncate({1}.mv_csr_{2}),2'b0}};\n".format(\
                        count, grp, csrname.lower())
                count += 1

                # cgfs are notoriously ordered and hence the following logic
                cfg_num = (num // (xlen//8))
                if xlen == 64 and (cfg_num%2) == 1:
                    cfg_num = cfg_num + 1
                cfg_offset = int(num%(xlen//8))
                cfg_grp = find_group(grp_yaml, 'PMPCFG'+str(cfg_num))
                pmpcfg_str += \
            '        lv_pmpcfg[{0}] = {1}.mv_csr_pmpcfg{2}[{3}:{4}];\n'.format(\
                    cfg_count, cfg_grp, cfg_num, cfg_offset*8+7,
                    cfg_offset*8)
                cfg_count += 1
        bsvstr += pmpaddr_def.format(pmp_entries, physical_addr_sz, pmpaddr_str)
        bsvstr += pmpcfg_def.format(pmp_entries, pmpcfg_str)

    # complete the value method for each csr.
    if logging:
        logger.debug('-- Defining value methods')
    bsvstr +='''
    interface Sbread sbread;'''
    for csrname in isa_yaml:
        if attr_yaml[csrname]['value_method']['required']:
            grp = find_group(grp_yaml, csrname.upper())
            bsvstr += textwrap.indent(methodv_def_temp.format(csrname, grp+'.mv_csr_', csrname), '  ')
    bsvstr +='''
    endinterface'''

    
    # define csr specific action methods.
    if logging:
        logger.debug('-- Defining custom action methods')
    for csrname in isa_yaml:
        logger.debug('Top CSR Interface: ' + csrname)
        if attr_yaml[csrname]['action_method']['required'] and \
                attr_yaml[csrname]['action_method']['global']: 
#            defs =  attr_yaml[csrname]['action_method']['def_arguments']
            csr_grp = find_group(grp_yaml, csrname.upper())
            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
                arguments = arguments.replace('xlen',str(xlen))
                arguments = arguments.replace('size',str(size))
                if '::' in arguments:
                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
                    if eval(cond):         
                        args = textwrap.indent(args, '')
                    else:
                        args = ''
                else:
                    args = arguments
                methods = re.findall(r'\s*method\s*Action\s*(.*?)[\s(]', args)
                for x in methods:
                    bsvstr += methoda_top_temp.format(x, csr_grp)
    # debug specific interfaces
    if debug_enabled:
        bsvstr += method_top_debug_def;
            

    # -- end the module
    bsvstr += end_module_temp
    # end package
    bsvstr += end_package_temp
    return bsvstr




import logging
import re
import sys
import os
import csrbox
import csrbox.utils as utils
from functools import reduce
from csrbox.constants import *
from itertools import chain 
import datetime
import textwrap

version = csrbox.__version__
logger = logging.getLogger(__name__)

def calculate_size(isa_yaml, custom_attr, logging=False):
    attr_yaml = utils.load_yaml(attr_file)
    if custom_attr:
        c_attr_yaml = utils.load_yaml(custom_attr)
    else:
        c_attr_yaml = {}
    rvxlen = 'rv' + str(isa_yaml['hart0']['supported_xlen'][0])
    std_subfields = standard_csr_fields.copy()
    std_subfields.remove('fields')
    for node in isa_yaml['hart0']:
        size = -1
        if isinstance(isa_yaml['hart0'][node] , dict) and \
                'address' in isa_yaml['hart0'][node]:
            if isa_yaml['hart0'][node][rvxlen]['accessible']:
                csrname = str(node).lower()
                if logging:
                    logger.debug('Calculating size for :' + str(csrname))
                if 'msb' in isa_yaml['hart0'][node][rvxlen]:
                    size = isa_yaml['hart0'][node][rvxlen]['msb']
                    if csrname in attr_yaml:
                        attr_yaml[csrname]['size'] = size + 1
                    elif csrname in c_attr_yaml:
                        c_attr_yaml[csrname]['size'] = size + 1
                        attr_yaml[csrname] = c_attr_yaml[csrname]
                    else:
                        attr_yaml[csrname] = {'size': size + 1,\
                                'value_method': {'required': True},
                                'action_method': {'required': False}}
                else:
                    for subfield, content in isa_yaml['hart0'][node][rvxlen].items():
                        if subfield not in std_subfields:
                            if subfield == 'fields':
                                for x in content:
                                    if isinstance(x,list):
                                        list_max = max([sublist[-1] for sublist in x])
                                        size = max(size, list_max)
                            else:
                                size = max(size, content['msb'])
                    if csrname in attr_yaml:
                        attr_yaml[csrname]['size'] = size + 1
                    elif csrname in c_attr_yaml:
                        c_attr_yaml[csrname]['size'] = size + 1
                        attr_yaml[csrname] = c_attr_yaml[csrname]
                    else:
                        attr_yaml[csrname] = {'size': size + 1,\
                                'value_method': {'required': True},
                                'action_method': {'required': False}}

    return attr_yaml

def csr_acc_false(bar):
   not_acc = []
   rvxlen = 'rv' + str(bar['hart0']['supported_xlen'][0])  
   for node in bar['hart0']:
            if isinstance(bar['hart0'][node], dict):
                if 'description' in bar['hart0'][node]:
                    csr_name = node
                    name_str = str(csr_name)
                    if bar['hart0'][csr_name][rvxlen]['accessible'] == False:
                       not_acc.append(name_str) 
   return not_acc 

def find_group(bar, csrname, logging=False):
    for grps in bar:
        if csrname.upper() in bar[grps]:
            return grps
    return None


def csrgrp_check(isa_yaml, grp_yaml, logging=False):
    if logging:
        logger.info('Sanity check for Grouping File')

    grp_list = list(grp_yaml)
    rvxlen = 'rv' + str(max(isa_yaml['hart0']['supported_xlen']))



    # - check for duplicates
    for grp, elem in grp_yaml.items():
      if elem != None :
        if len(elem) != len(set(elem)) :
            dup = list(set([x for x in elem if elem.count(x) > 1]))
            logger.error( grp + ' has the following duplicates ' + str(dup))
            raise SystemExit

    # - check of overlaps
    for i in range(0,len(grp_list)-1):
        x = grp_yaml[grp_list[i]]
        for j in range(i+1, len(grp_list)):
            y = grp_yaml[grp_list[j]]
            if x != None and y != None:
             intersection = list(set(x) & set(y))
            if intersection:
                logger.error(grp_list[i] + ' and ' + grp_list[j] + 
                        ' have the following common entries '+ str(intersection))
                raise SystemExit
    if rvxlen == 'rv64':
        same_groups = same_groups_rv64
    else:
        same_groups = same_groups_rv32

    # - same group checks
    for grp, elem in grp_yaml.items():
      if elem != None :
        for csrs in elem:
            for pairs in same_groups:
                if csrs in pairs:
                    grp1 = find_group(grp_yaml, pairs[0])
                    grp2 = find_group(grp_yaml, pairs[1])
                    if grp1 is not None and grp2 is not None and grp1 != grp2:
                        logger.error(str(pairs) + ' Need to be in same group')
                        raise SystemExit
            if rvxlen == 'rv32':
                for pairs in necessary_pairs_rv32:
                    if csrs in pairs:
                        grp1 = find_group(grp_yaml, pairs[0])
                        grp2 = find_group(grp_yaml, pairs[1])
                        if grp1 is None or grp2 is None:
                            logger.error('Please instantiate both the following pairs: ' + str(pairs))
                            raise SystemExit
    # -- grp naming required
    for i in range(1, len(grp_list)+1):
        if 'grp'+str(i) not in grp_yaml:
            logger.error('grp'+str(i) + ' is missing')
            raise SystemExit

    # -- grp csrs must be accessible
    for grp, elems in grp_yaml.items():
      if elem != None :
        for csr in elems:
            if csr.lower() not in isa_yaml['hart0']:
                logger.error(csr + ' in grouping YAML not found in ISA-YAML')
                raise SystemExit
            if not isa_yaml['hart0'][csr.lower()][rvxlen]['accessible']:
                logger.error( csr + ' in grouping yaml is not accessible in ISA=YAML')
                raise SystemExit

    # -- missing csrs in grp_yaml
    missing = []
    for csr, elem in isa_yaml['hart0'].items():
        if isinstance(elem, dict):
            if 'description' in elem and elem[rvxlen]['accessible']:
                if find_group(grp_yaml, csr) is None:
                    missing.append(csr)
    if missing:
        logger.warn('The following CSRS are accessible in the ISA-YAML but missing the grouping yaml' + str(missing))

    logger.info('Grouping Yaml is valid')

